__author__ = "Denis V. Shakalo"
__copyright__ = "Copyright (C) 2018 Denis V. Shakalo"
__license__ = """Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/COPYING-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
__version__ = "0.0.2"

import subprocess
import re
import socket
import fcntl
import struct
import os
from subprocess import Popen, PIPE
from kvipset.vipseterror import VipsetError



class System:
    """Common functions"""

    @staticmethod
    def get_default_ifname():
        """Return default route interface name"""
        with open('/proc/net/route', 'r') as routes:
            for line in routes:
                parsed_line = line.split('\t')
                if parsed_line[1] == '00000000':
                    return parsed_line[0]
        return False

    @staticmethod
    def is_valid_ip(ip,prefix):
        """Validate IP"""
        ip_regex = re.compile(
            '^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])(/(3[0-2]|[1-2][0-9]|[0-9]))$')
        fullvip = '{}/{}'.format(ip, prefix)
        return ip_regex.match(fullvip)

    @staticmethod
    def check_ifname(ifname):
        """Check if configured ifname exists in the system"""
        pattern = re.compile('^\s*'+ifname+':')
        with open('/proc/net/dev', 'r') as net_devs:
            for line in net_devs:
                if pattern.match(line):
                    return True
        return False

    @staticmethod
    def run_cmd(cmd, cmdtag):
        """Run shell command"""
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True)
            out, err = p.communicate()
            err = err.decode('utf8')
            status = p.wait()
        except Exception as e:
            raise VipsetError(cmd, e.args, cmdtag)
        if not os.WIFEXITED(status) or os.WEXITSTATUS(status):
            if not re.match(r'.*File exists.*', err):
                raise VipsetError(cmd, err, cmdtag)
        return str(out, 'utf-8')

    @staticmethod
    def add_vip(vip, prefix, ifname):
        fullvip = '{}/{}'.format(vip, prefix)
        cmd = ['ip', 'addr', 'add', fullvip, 'dev', ifname]
        System.run_cmd(cmd, 'ipadd')

    @staticmethod
    def del_vip(vip, prefix, ifname):
        fullvip = '{}/{}'.format(vip, prefix)
        cmd = ['ip', 'addr', 'delete', fullvip, 'dev', ifname]
        System.run_cmd(cmd, 'ipdel')

    @staticmethod
    def get_primary_ip_address(ifname):
        """ Get interface primary IP address"""
        try:
            with (socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
                ip = socket.inet_ntoa(fcntl.ioctl(
                s.fileno(),
                0x8915,  # SIOCGIFADDR
                struct.pack(b'256s', ifname[:15].encode('utf-8'))
                )[20:24])
        except Exception as e:
            raise VipsetError('socket.inet_nto', e.args, 'getprimaryip')
        return ip

    @staticmethod
    def add_vip_bundle(vip, real_ip, ifname):
        """Add vip and arp filter"""
        if System.is_valid_ip(vip['ipaddr'], vip['prefix']):
            try:
                if not Arptables.find_arp_filter(vip['ipaddr'], real_ip):
                    # Check if arp filter already exists
                    print(
                        'Trying to configure arp filter for VIP {} and RealIP {}'.format(vip['ipaddr'], real_ip))
                    Arptables.add_arp_filter(vip['ipaddr'], real_ip)
                    # Set filter
                else:
                    print('Arp filter for VIP {} and RealIP {} already exists'.format(vip['ipaddr'], real_ip))
                if Arptables.find_arp_filter(vip['ipaddr'], real_ip):
                    # Check if filter has been configured
                    print(
                        'Arp filter for VIP {} and RealIP {} has been configured'.format(vip['ipaddr'], real_ip))
                    System.add_vip(vip['ipaddr'], vip['prefix'], ifname)
                    # Add VIP
                    print('VIP {}/{} has been configured'.format(vip['ipaddr'], vip['prefix']))
            except VipsetError as e:
                print(e.command, e.message, e.cmdtag)
        else:
            print('{} is invalid'.format(vip))

    @staticmethod
    def del_vip_bundle(vip, real_ip, ifname):
        """Remove vip and arp filter"""
        if System.is_valid_ip(vip['ipaddr'], vip['prefix']):
            try:
                System.del_vip(vip['ipaddr'], vip['prefix'], ifname)
                # Try to delete VIP
                print('VIP {}/{} has been removed'.format(vip['ipaddr'], vip['prefix']))
            except VipsetError as e:
                print(e.command, e.message, e.cmdtag)
            try:
                if Arptables.find_arp_filter(vip['ipaddr'], real_ip):
                    # Check if arp filter exists
                    print('Trying to remove arp filter for VIP {} and RealIP {}'.format(vip['ipaddr'], real_ip))
                    Arptables.del_arp_filter(vip['ipaddr'], real_ip)
                    # Remove arp filter
                if not Arptables.find_arp_filter(vip['ipaddr'], real_ip):
                    print('Arp filter for VIP {} and RealIP {} has been removed'.format(vip['ipaddr'], real_ip))
            except VipsetError as e:
                print(e.command, e.message, e.cmdtag)
        else:
            print('{} is invalid'.format(vip))

    @staticmethod
    def get_if_ips(ifname):
        """Parse ip addr output to find all interface ip addresses"""
        ip_regex = re.compile(
            '(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])(/(3[0-2]|[1-2][0-9]|[0-9]))')
        ips = set()
        for ip in ip_regex.finditer(System.run_cmd(['ip', '-o', 'addr', 'show', 'dev', ifname], 'ipaddr')):
            ips.add(ip.group())
        return ips

    @staticmethod
    def print_exception(e):
        """Print exception error message"""
        if e.cmdtag == 'configgetdefaultifname':
            print('DSR interface is unconfigured and unable to get default interface')
        else:
            print('Unknown error on {}'.format(e.cmdtag))

class Arptables:

    @staticmethod
    def add_arp_filter(vip, real_ip):
        """Add arp filtering record pair"""
        System.run_cmd(['arptables', '-A', 'INPUT', '-j', 'DROP', '-d', vip],'arpdropadd')
        System.run_cmd(['arptables', '-A', 'OUTPUT', '-j', 'mangle', '-s', vip, '--mangle-ip-s', real_ip],'arpmangleadd')

    @staticmethod
    def del_arp_filter(vip, real_ip):
        """Delete arp filtering pair"""
        System.run_cmd(['arptables', '-D', 'INPUT', '-j', 'DROP', '-d', vip],'arpdropadd')
        System.run_cmd(['arptables', '-D', 'OUTPUT', '-j', 'mangle', '-s', vip, '--mangle-ip-s', real_ip],'arpmangleadd')

    @staticmethod
    def find_arp_filter(vip, real_ip):
        """Returns true if correct arp filter for vip-real_ip pair is configured"""
        p = Popen(['arptables', '-nL'], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = p.communicate()
        if not p.returncode:
            if output.find(b'-j DROP -d '+bytearray(vip,'utf-8')) != -1:
                if output.find(b'-j mangle -s '+bytearray(vip,'utf-8')+b' --mangle-ip-s '+bytearray(real_ip,'utf-8')) != -1:
                    return True
        return False
